def get_empty_znode(node_id):
    """
    Get an empty ZNode with headers filled.

    Args:
        node_id: String that identifies the ZNode

    Returns:
        A dictionary representing a ZNRecord
    """
    return {
        'id': node_id, 'simpleFields': {}, 'listFields': {}, 'mapFields': {}}



import pyhelix.statemodel as statemodel


class DummyStateModel(statemodel.StateModel):
    """
    Dummy state model for master-slave
    """

    def on_become_slave_from_offline(self, message):
        partition_name = message['simpleFields']['PARTITION_NAME']
        print('{0}: on become slave from offline'.format(partition_name))

    def on_become_offline_from_slave(self, message):
        partition_name = message['simpleFields']['PARTITION_NAME']
        print('{0}: on become offline from slave'.format(partition_name))

    def on_become_master_from_slave(self, message):
        partition_name = message['simpleFields']['PARTITION_NAME']
        print('{0}: on become master from slave'.format(partition_name))

    def on_become_slave_from_master(self, message):
        partition_name = message['simpleFields']['PARTITION_NAME']
        print('{0}: on become slave from master'.format(partition_name))

    def on_become_dropped_from_offline(self, message):
        partition_name = message['simpleFields']['PARTITION_NAME']
        print('{0}: on become dropped from offline'.format(partition_name))


class DummyStateModelFactory(statemodel.StateModelFactory):
    """
    Dummy state model factory
    """

    def create_state_model(self, partition_name):
        """
        Create a DummyStateModel

        Args:
            partition_name: The partition

        Returns:
            DummyStateModel
        """
        state_model = DummyStateModel()
        return state_model



import unittest

import pyhelix.statemodel as statemodel


class MockStateModel(statemodel.StateModel):
    existing_invoked = False
    default_invoked = False

    def on_become_online_from_offline(self, message):
        self.existing_invoked = True

    def default_transition_handler(self, message):
        self.default_invoked = True


class TestStateModelParser(unittest.TestCase):
    """
    Make sure that functions are invoked properly based on transition messages.
    """

    def setUp(self):
        self._state_model = MockStateModel()
        self._parser = statemodel.StateModelParser()

    def test_normal(self):
        """
        Normal case: all caps states invokes defined method
        """
        method = self._parser.get_method_for_transition(
            self._state_model, 'OFFLINE', 'ONLINE')
        method(None)
        self.assertTrue(self._state_model.existing_invoked)

    def test_default(self):
        """
        Defined method doesn't exist, so invoke the default
        """
        method = self._parser.get_method_for_transition(
            self._state_model, 'GOOFY', 'SILLY')
        method(None)
        self.assertTrue(self._state_model.default_invoked)

    def test_case_insensitive(self):
        """
        Invoke the correct method even when casing is weird
        """
        method = self._parser.get_method_for_transition(
            self._state_model, 'offliNE', 'onLIne')
        method(None)
        self.assertTrue(self._state_model.existing_invoked)



import kazoo.client
import logging
import os

import accessor
import constants
import helixexec
import znode


class Participant(object):
    """
    Participant Helix connection

    This class encompasses all of a Helix participant's interactions with
    ZooKeeper.
    """
    def __init__(self, cluster_id, host, port, zk_addrs, participant_id=None):
        """
        Initialize the connection parameters.

        Args:
            cluster_id: The cluster this participant should belong to
            host: Host of this participant
            port: Logical port of this participant
            zk_addrs: Comma separated host:port of ZooKeeper servers
            participant_id: (Optional) Custom ID, "host_port" by default
        """
        self._host = host
        self._port = port
        if participant_id:
            self._participant_id = participant_id
        else:
            self._participant_id = '{0}_{1}'.format(host, port)
        self._client = kazoo.client.KazooClient(zk_addrs)
        self._client.add_listener(self._connection_listener)
        self._accessor = accessor.DataAccessor(cluster_id, self._client)
        self._builder = self._accessor.get_key_builder()
        self._callbacks = set()
        self._state_model_ftys = {}
        self._executor = helixexec.HelixExecutor(self._state_model_ftys, self)
        self._pre_connect_callbacks = set()
        self._is_lost = False

    def connect(self):
        """
        Establish a connection.
        """
        # Connect to ZK
        self._is_lost = False
        self._client.start()
        self._init()

    def disconnect(self):
        """
        End an active connection.
        """
        self._is_lost = True
        self._accessor.remove(
            self._builder.live_instance(self._participant_id))
        self._client.stop()
        self._client.close()

    def is_connected(self):
        """
        Get the current connection status.

        Returns:
            True if connected, False if disconnected
        """
        return self._client.connected

    def get_accessor(self):
        """
        Get a DataAccessor for this cluster.

        Returns:
            Instantiated DataAccessor
        """
        return self._accessor

    def get_participant_id(self):
        """
        Get the ID of this participant

        Returns:
            String participant ID
        """
        return self._participant_id

    def get_session_id(self):
        """
        Get the ID of the current ZooKeeper session

        Returns:
            Session ID, or None if not connected
        """
        if not self.is_connected():
            logging.error('Cannot get session ID, not connected')
            return None
        return str(self._client.client_id[0])

    def register_pre_connect_callback(self, callback):
        """
        Add a callback that will be invoked just prior to joining.

        Args:
            callback: A function that takes no arguments
        """
        self._pre_connect_callbacks.add(callback)

    def register_state_model_fty(self, state_model_name, state_model_fty):
        """
        Register a state model factory with transition callbacks

        Args:
            state_model_name: The name of the state model definition
            state_model_fty: A factory of callbacks for state transitions
        """
        self._state_model_ftys[state_model_name] = state_model_fty

    def unregister_state_model_fty(self, state_model_name):
        """
        Unregister a state model factory.

        Args:
            state_model_name: The state model for which to ignore transitions
        """
        self._state_model_ftys.pop(state_model_name)

    def _register_message_callback(self, callback):
        """
        Register a callback for messages to this participant (private)

        Args:
            callback: A function that takes a list of message IDs.
        """
        self._callbacks.add(callback)

    def _ensure_participant_config(self):
        """
        Ensure that ZNodes for a participant all exist, non-atomic (private)

        Returns:
            True if everything was persisted, False otherwise
        """
        exists = self._accessor.exists(
            self._builder.participant_config(self._participant_id))
        if not exists and self._auto_join_allowed():
            node = znode.get_empty_znode(self._participant_id)
            node['simpleFields'] = {
                'HELIX_HOST': self._host, 'HELIX_PORT': str(self._port),
                'HELIX_ENABLED': 'true'}
            self._accessor.create(
                self._builder.participant_config(self._participant_id), node)
            self._accessor.create(
                self._builder.instance(self._participant_id), b'')
            self._accessor.create(
                self._builder.current_states(self._participant_id), b'')
            self._accessor.create(
                self._builder.errors(self._participant_id), b'')
            self._accessor.create(
                self._builder.health_report(self._participant_id), b'')
            self._accessor.create(
                self._builder.messages(self._participant_id), b'')
            self._accessor.create(
                self._builder.status_updates(self._participant_id), b'')
        elif not exists:
            return False
        return True

    def _auto_join_allowed(self):
        """
        Check if this cluster supports participants auto-joining (private)

        Returns:
            True if auto join allowed, False otherwise
        """
        clst_config = self._accessor.get(self._builder.cluster_config())
        if clst_config:
            if 'allowParticipantAutoJoin' in clst_config['simpleFields']:
                auto_join = (
                    clst_config['simpleFields']['allowParticipantAutoJoin'])
                if auto_join == 'true':
                    return True
        return False

    def _create_live_instance_node(self):
        """
        Create a live instance ZNode for this participant (private)

        Returns:
            True if created, False otherwise
        """
        node = znode.get_empty_znode(self._participant_id)
        node['simpleFields'] = {
            'HELIX_VERSION': 'pyhelix-{0}'.format(constants.CURRENT_VERSION),
            'SESSION_ID': str(self._client.client_id[0]),
            'LIVE_INSTANCE': '{0}@{1}'.format(os.getpid(), self._host)}
        return self._accessor.create(
            self._builder.live_instance(self._participant_id), node)

    def _message_handler(self, messages):
        """
        Dispatcher for message callbacks (private)

        Args:
            List of message ids as strings

        Returns:
            Always True
        """
        logging.info('handler called: {0}'.format(messages))
        for cb in self._callbacks:
            message_nodes = []
            for message_id in messages:
                message_nodes.append(self._accessor.get(
                    self._builder.message(self._participant_id, message_id)))
            cb(message_nodes)
        return True

    def _connection_listener(self, state):
        """
        Callback for connection state changes (private)

        Args:
            state: the current connection state (LOST, CONNECTED, SUSPENDED)
        """
        if state == kazoo.client.KazooState.LOST:
            self._is_lost = True
        elif self._is_lost and state == kazoo.client.KazooState.CONNECTED:
            self._client.handler.spawn(self._init)
            self._is_lost = False

    def _init(self):
        """
        Internal initialization (private)
        """
        # Register with the cluster
        result = self._ensure_participant_config()
        if not result:
            logging.error('Participant configuration could not be added')
            self.disconnect()
            return

        # Get ready to receive cluster messages
        self._register_message_callback(self._executor.on_message)
        self._accessor.watch_children(self._builder.messages(
            self._participant_id), self._message_handler)

        # Invoke pre-connect callbacks
        for pre_connect_callback in self._pre_connect_callbacks:
            pre_connect_callback()

        # Create an ephemeral node
        result = self._create_live_instance_node()
        if not result:
            logging.error('Could not create live instance')
            self.disconnect()
            return



import kazoo.client
import kazoo.exceptions


class MockStruct(object):
    """
    Basic object to allow adding arbitrary properties
    """


class MockKazooClient(object):
    """
    In-memory version of the kazoo client for unit testing
    """

    def __init__(self):
        self.store = {'/': None}
        self.ephemerals = set()
        self.versions = {}
        self._connected = False

    def start(self):
        self._connected = True

    def stop(self):
        for eph in self.ephemerals:
            self.store.pop(eph)
        self._connected = False

    @property
    def connected(self):
        return self._connected

    @property
    def client_id(self):
        return 123, 'password'

    def restart(self):
        self.stop()
        self.start()

    def close(self):
        pass

    def create(self, path, data, ephemeral=False, sequence=False,
               makepath=False):
        # TODO: sequence not implemented
        if path in self.store or path == '/':
            raise kazoo.exceptions.NodeExistsError
        parts = path.split('/')
        for i in xrange(1, len(parts)):
            subpath = '/'.join(parts[:i])
            if not subpath:
                subpath = '/'
            if subpath in self.store and subpath in self.ephemerals:
                raise kazoo.exceptions.NoChildrenForEphemeralsError
            if subpath not in self.store:
                if makepath or subpath == '/':
                        self.store[subpath] = None
                else:
                    raise kazoo.exceptions.NoNodeError
        if ephemeral:
            self.ephemerals.add(path)
        self.store[path] = data
        return path

    def ensure_path(self, path):
        if path in self.store or path == '/':
            return
        parts = path.split('/')
        for i in xrange(1, len(parts)):
            subpath = '/'.join(parts[:i])
            if subpath not in self.store:
                self.store[subpath] = None
        self.store[path] = None

    def exists(self, path):
        return path if path in self.store else None

    def get(self, path):
        # TODO: manage versions internally better
        if path not in self.store:
            raise kazoo.exceptions.NoNodeError
        get_stat = MockStruct()
        get_stat.version = -1
        return self.store[path], True

    def get_children(self, path, include_data=False):
        # TODO: include_data doesn't do the right thing
        if path not in self.store:
            raise kazoo.exceptions.NoNodeError
        parts = path.split('/')
        result = []
        for existpath, data in self.store.iteritems():
            if existpath.startswith(path):
                existparts = existpath.split('/')
                if len(existparts) == len(parts) + 1:
                    if include_data:
                        result.append((existpath, data))
                    else:
                        result.append(existpath)
        return result

    def set(self, path, data, version=-1):
        if path not in self.store:
            raise kazoo.exceptions.NoNodeError
        if path in self.versions:
            existversion = self.versions[path]
            if version != -1 and existversion > version:
                raise kazoo.exceptions.BadVersionError
        self.store[path] = data
        self.versions[path] = version

    def delete(self, path, version=-1, recursive=False):
        if path not in self.store:
            raise kazoo.exceptions.NoNodeError
        existversion = -1
        if path in self.versions:
            existversion = self.versions[path]
            if version != -1 and existversion > version:
                raise kazoo.exceptions.BadVersionError
        to_pop = []
        for existpath in self.store.iterkeys():
            if existpath.startswith(path):
                if path != existpath and not recursive:
                    raise kazoo.exceptions.NotEmptyError
                to_pop.append(existpath)
        for subpath in to_pop:
            self.store.pop(subpath)
            if subpath in self.versions:
                self.versions.pop(subpath)
            if subpath in self.ephemerals:
                self.ephemerals.pop(subpath)
        if path == '/':
            self.store = {'/': None}

    def add_listener(self, unused):
        pass

    def remove_listener(self, unused):
        pass



import pyhelix.statemodel as statemodel


class CodeRunnerModel(statemodel.StateModel):
    """
    State model for an online-offline code runner set
    """
    def __init__(self):
        """
        Initialize a CodeRunnerModel
        """
        self._active = False

    def is_active(self):
        """
        Check if requests should be actively served

        Returns:
            True if online, False if offline
        """
        return self._active

    def on_become_online_from_offline(self, message):
        partition_name = message['simpleFields']['PARTITION_NAME']
        self._active = True
        print('{0}: on become online from offline'.format(partition_name))

    def on_become_offline_from_online(self, message):
        partition_name = message['simpleFields']['PARTITION_NAME']
        self._active = False
        print('{0}: on become offline from online'.format(partition_name))

    def on_become_dropped_from_offline(self, message):
        partition_name = message['simpleFields']['PARTITION_NAME']
        print('{0}: on become dropped from offline'.format(partition_name))

    def on_become_offline_from_error(self, message):
        partition_name = message['simpleFields']['PARTITION_NAME']
        self._active = False
        print('{0}: on become offline from error'.format(partition_name))


class CodeRunnerModelFactory(statemodel.StateModelFactory):
    """
    State model factory for an online-offline code runner set
    """
    def create_state_model(self, partition_name):
        """
        Create a CodeRunnerModel

        Args:
            partition_name: The partition

        Returns:
            CodeRunnerModel
        """
        return CodeRunnerModel()



import json
import kazoo.client
import logging

import accessor


class SpectatorConnection(object):
    """
    Helix spectator connection

    This class encompasses all of a spectator's interactions with ZooKeeper.
    """
    def __init__(self, cluster_id, zk_addrs):
        """
        Initialize the connection parameters

        Args:
            cluster_id: The cluster to watch
            zk_addrs: Comma-separated host:port of ZooKeeper servers
        """
        self._cluster_id = cluster_id
        self._client = kazoo.client.KazooClient(zk_addrs)
        self._client.add_listener(self._connection_listener)
        self._accessor = accessor.DataAccessor(cluster_id, self._client)
        self._keybuilder = self._accessor.get_key_builder()
        self._spectators = {}
        self._participants = {}
        self._is_lost = False

    def connect(self):
        """
        Establish a connection
        """
        self._is_lost = False
        self._client.start()
        self._init()

    def disconnect(self):
        """
        End an active connection
        """
        self._is_lost = True
        self._client.stop()

    def is_connected(self):
        """
        Get the current connection status

        Returns:
            True if connected, False otherwise
        """
        return self._client.connected

    def spectate(self, resource_id):
        """
        Start spectating on a resource.

        Args:
            resource_id: The resource of interest

        Returns:
            A Spectator object
        """
        if not self.is_connected():
            logging.error(
                'Tried to spectate on {0} without connecting!'.format(
                    resource_id))
            return None
        if resource_id in self._spectators:
            return self._spectators[resource_id]
        logging.debug('About to start watching {0}'.format(resource_id))
        s = Spectator(self._accessor, resource_id, self._participants)
        self._spectators[resource_id] = s
        return s

    def get_accessor(self):
        """
        Get a DataAccessor for this cluster.

        Returns:
            Instantiated DataAccessor
        """
        return self._accessor

    def _pc_parent_watcher(self, children):
        """
        Callback for participant config children (private)

        Args:
            children: List of child names

        Returns:
            Always True
        """
        if not children:
            return True
        for child in children:
            if child not in self._participants:
                # Watch each participant config
                self._accessor.watch_property(
                    self._keybuilder.participant_config(child),
                    self._pc_watcher)
        return True

    def _pc_watcher(self, data, stat):
        """
        Callback for participant config change (private)

        Args:
            data: The ZNode data
            stat: The ZNode stat

        Returns:
            Always True
        """
        if not data:
            return True
        participant_config = json.loads(data)
        if participant_config and 'id' in participant_config:
            self._participants[participant_config['id']] = participant_config
        return True

    def _connection_listener(self, state):
        """
        Callback for connection state changes (private)

        Args:
            state: the current connection state (LOST, CONNECTED, SUSPENDED)
        """
        if state == kazoo.client.KazooState.LOST:
            self._is_lost = True
        elif self._is_lost and state == kazoo.client.KazooState.CONNECTED:
            self._client.handler.spawn(self._init)
            self._is_lost = False

    def _init(self):
        """
        Internal initialization (private)
        """
        self._participants.clear()
        self._accessor.watch_children(
            self._keybuilder.participant_configs(), self._pc_parent_watcher)
        for resource_id, s in self._spectators.iteritems():
            s._init(resource_id)


class Spectator(object):
    """
    Helix spectator
    """
    def __init__(self, accessor, resource_id, participants):
        """
        Initialize a spectator for a resource

        Args:
            accessor: Instantiated DataAccessor
            resource_id: The resource to spectate
            participants: Map of participant id to participant config
        """
        self._mapping = {}
        self._participants = participants
        self._accessor = accessor
        self._keybuilder = accessor.get_key_builder()
        self._init(resource_id)

    def get_participants(self, state, partition_id=None):
        """
        Get all participants in a given state (optionally for a partition)

        Args:
            state: The state
            partition_id: The partition (optional)

        Returns:
            List of participants
        """
        result = set()
        partitions = None
        if partition_id:
            partitions = [partition_id]
        else:
            partitions = self._mapping.keys()
        for partition_id in partitions:
            for participant_id, s in self._mapping[partition_id].iteritems():
                if s == state:
                    result.add(participant_id)
        return [self._participants[p] for p in result]

    def get_state_map(self, partition_id):
        """
        Get a mapping of participant to state for a partition

        Args:
            partition_id: The partition

        Returns:
            Map of participant id to state
        """
        if partition_id in self._mapping:
            return self._mapping[partition_id]
        else:
            return {}

    def _ev_watcher(self, data, stat):
        """
        Callback for external view change (private)

        Args:
            data: The ZNode data
            stat: The ZNode metadata

        Returns:
            Always True
        """
        if not data:
            self._mapping = {}
            return True
        external_view = json.loads(data)
        if (external_view and
           'mapFields' in external_view and external_view['mapFields']):
            self._mapping = external_view['mapFields']
        else:
            self._mapping = {}
        logging.debug('Updated external view: {0}'.format(self._mapping))
        return True

    def _init(self, resource_id):
        """
        Internal initialization (private)

        Args:
            resource_id: The resource to spectate on
        """
        self._accessor.watch_property(
            self._keybuilder.external_view(resource_id), self._ev_watcher)



def propertykey(ephemeral=False, sequential=False, merge_on_update=False,
                update_only_on_exists=False):
    def decorator(func):
        def hidden_func(*args, **kwargs):
            prop = {'ephemeral': ephemeral,
                    'sequential': sequential,
                    'merge_on_update': merge_on_update,
                    'update_only_on_exists': update_only_on_exists}
            prop['path'] = func(*args, **kwargs)
            return prop
        return hidden_func
    return decorator


class KeyBuilder(object):
    """
    A collection of functions to get ZooKeeper metadata
    """
    def __init__(self, cluster_id):
        self._cluster_id = cluster_id

    @propertykey()
    def cluster_config(self):
        return '/{0}/CONFIGS/CLUSTER/{0}'.format(self._cluster_id)

    @propertykey()
    def external_views(self):
        return '/{0}/EXTERNALVIEW'.format(self._cluster_id)

    @propertykey()
    def external_view(self, resource_id):
        return '/{0}/EXTERNALVIEW/{1}'.format(self._cluster_id, resource_id)

    @propertykey()
    def ideal_states(self, resource_id):
        return '/{0}/IDEALSTATES'.format(self._cluster_id)

    @propertykey()
    def ideal_state(self, resource_id):
        return '/{0}/IDEALSTATES/{1}'.format(self._cluster_id, resource_id)

    @propertykey()
    def participant_configs(self):
        return '/{0}/CONFIGS/PARTICIPANT'.format(self._cluster_id)

    @propertykey()
    def participant_config(self, participant_id):
        return '/{0}/CONFIGS/PARTICIPANT/{1}'.format(
            self._cluster_id, participant_id)

    @propertykey()
    def live_instances(self):
        return '/{0}/LIVEINSTANCES'.format(self._cluster_id)

    @propertykey()
    def instance(self, participant_id):
        return '/{0}/INSTANCES/{1}'.format(self._cluster_id, participant_id)

    @propertykey(ephemeral=True)
    def live_instance(self, participant_id):
        return '/{0}/LIVEINSTANCES/{1}'.format(
            self._cluster_id, participant_id)

    @propertykey()
    def current_states(self, participant_id, session_id=None):
        path = '/{0}/INSTANCES/{1}/CURRENTSTATES'.format(
            self._cluster_id, participant_id, session_id)
        if session_id:
            path += '/{0}'.format(session_id)
        return path

    @propertykey(merge_on_update=True)
    def current_state(self, participant_id, session_id, resource_id):
        return '/{0}/INSTANCES/{1}/CURRENTSTATES/{2}/{3}'.format(
            self._cluster_id, participant_id, session_id, resource_id)

    @propertykey(merge_on_update=True)
    def errors(self, participant_id):
        return '/{0}/INSTANCES/{1}/ERRORS'.format(
            self._cluster_id, participant_id)

    @propertykey(merge_on_update=True)
    def health_report(self, participant_id):
        return '/{0}/INSTANCES/{1}/HEALTHREPORT'.format(
            self._cluster_id, participant_id)

    @propertykey()
    def status_updates(self, participant_id):
        return '/{0}/INSTANCES/{1}/STATUSUPDATES'.format(
            self._cluster_id, participant_id)

    @propertykey()
    def state_models(self):
        return '/{0}/STATEMODELDEFS'.format(self._cluster_id)

    @propertykey()
    def state_model(self, resource_id):
        return '/{0}/STATEMODELDEFS/{1}'.format(self._cluster_id, resource_id)

    @propertykey()
    def messages(self, participant_id):
        return '/{0}/INSTANCES/{1}/MESSAGES'.format(
            self._cluster_id, participant_id)

    @propertykey(merge_on_update=True, update_only_on_exists=True)
    def message(self, participant_id, message_id):
        return '/{0}/INSTANCES/{1}/MESSAGES/{2}'.format(
            self._cluster_id, participant_id, message_id)



import argparse
import bottle
import cStringIO
import logging
import sys
import time
import traceback

import pyhelix.participant as participant

import runnermodel

logging.basicConfig(level=logging.WARN)


class CodeRunnerProcess(object):
    """
    A class that makes a node that runs code available
    """
    def __init__(self, cluster, host, port, zk_svr):
        """
        Initialize the CodeRunnerProcess

        Args:
            cluster: the cluster id
            host: the hostname that this runner can be reached on
            port: the port to use to send data
            zk_svr: host:port of a ZooKeeper server
        """
        self._p = participant.Participant(cluster, host, port, zk_svr)
        self._factory = runnermodel.CodeRunnerModelFactory()
        self._p.register_state_model_fty(
            'OnlineOfflineModified', self._factory)
        self._app = bottle.Bottle()
        self._host = host
        self._port = port
        self._route()

    def start(self):
        """
        Start the CodeRunnerProcess
        """
        self._p.connect()
        self._app.run(host='0.0.0.0', port=self._port)

    def stop(self):
        """
        Stop the CodeRunnerProcess
        """
        self._p.disconnect()

    def run_program(self):
        """
        Called when a request with a program to run comes in

        Returns:
            stdout result of running the program
        """
        logging.info('New request to {0}:{1}'.format(self._host, self._port))
        bottle.response.add_header('Content-Type', 'text/plain')
        if not self._can_serve():
            return '[ERROR] Not accepting program submissions\n'
        prog = str(bottle.request.forms.get('prog'))
        old_stdout = sys.stdout
        redirected_output = sys.stdout = cStringIO.StringIO()
        try:
            exec(prog)
        except:
            print traceback.format_exc()
        sys.stdout = old_stdout
        return redirected_output.getvalue()

    def _can_serve(self):
        """
        Check if this machine should serve requests

        Returns:
            True if requests are accepted, False otherwise
        """
        model = self._factory.get_state_model('coderunner_0')
        if model.is_active():
            return True
        else:
            return False

    def _route(self):
        """
        Route HTTP requests (private)
        """
        self._app.route('/run', method='POST', callback=self.run_program)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--zkSvr', required=True, type=str, help='host:port of ZooKeeper')
    parser.add_argument('--host', required=True, type=str, help='hostname')
    parser.add_argument('--port', required=True, type=str, help='port')
    args = parser.parse_args()
    r = CodeRunnerProcess(
        'coderunner-cluster', args.host, args.port, args.zkSvr)
    r.start()
    r.stop()



import unittest

import pyhelix.participant as participant


class TestParticipant(unittest.TestCase):
    """
    These test methods test various parts of the participant lifecycle
    """

    def test_derived_id(self):
        """
        Test that a participant id is derived when none is specified
        """
        host = 'localhost'
        port = 123
        p = participant.Participant(
            'test-cluster', host, port, 'localhost:2181')
        self.assertEqual(p._participant_id, '{0}_{1}'.format(host, port))

    def test_custom_id(self):
        """
        Test that a custom participant id is honored
        """
        host = 'localhost'
        port = 123
        participant_id = 'myid'
        p = participant.Participant(
            'test-cluster', host, port, 'localhost:2181',
            participant_id=participant_id)
        self.assertEqual(p._participant_id, participant_id)



import concurrent.futures as futures
import logging
import time

import helixtask
import znode


class HelixExecutor(object):
    """
    Helix executor that listens on transition messages and schedules tasks
    """

    DEFAULT_PARALLELISM = 20

    def __init__(self, state_model_ftys, participant, num_concurrent=None):
        """
        Initialize the executor

        Args:
            state_model_ftys: An iterable collection of state model factories
            participant: A Helix participant object
        """
        if not num_concurrent:
            num_concurrent = self.DEFAULT_PARALLELISM
        self._threadpool = futures.ThreadPoolExecutor(num_concurrent)
        self._state_model_ftys = state_model_ftys
        self._accessor = participant.get_accessor()
        self._builder = self._accessor.get_key_builder()
        self._participant_id = participant.get_participant_id()
        self._participant = participant

    def on_message(self, messages):
        """
        Process incoming messages.

        Args:
            messages: Message ZNodes
        """
        for message in messages:
            # Skip bad messages
            if (not message or
               not message['simpleFields'] or
               not message['simpleFields']['MSG_TYPE']):
                continue

            # Skip messages that aren't transition messages
            message_type = message['simpleFields']['MSG_TYPE'].upper()
            if message_type != 'STATE_TRANSITION':
                continue

            # Remove messages that aren't for this session
            tgt_session_id = message['simpleFields']['TGT_SESSION_ID']
            session_id = self._participant.get_session_id()
            if tgt_session_id != session_id:
                logging.warn(
                    'Message {0} has target session id {1},'
                    ' expected {2}'.format(
                        message['id'], tgt_session_id, session_id))
                self._accessor.remove(
                    self._builder.message(self._participant_id, message['id']))
                continue

            # Skip messages that are already read
            message_state = message['simpleFields']['MSG_STATE'].upper()
            if message_state != 'NEW':
                continue

            # Get the state model, instantiating if it doesn't exist
            state_model_name = message['simpleFields']['STATE_MODEL_DEF']
            state_model_fty = self._state_model_ftys[state_model_name]
            partition_name = message['simpleFields']['PARTITION_NAME']
            state_model = state_model_fty.get_state_model(partition_name)
            if state_model is None:
                state_model = state_model_fty.create_state_model(
                    partition_name)
                state_model_fty.put_state_model(partition_name, state_model)

            # Update message to READ
            message['simpleFields']['MSG_STATE'] = 'READ'
            message['simpleFields']['READ_TIMESTAMP'] = '{0}'.format(
                int(time.time() * 1000))
            message['simpleFields']['EXE_SESSION_ID'] = session_id
            self._accessor.update(self._builder.message(
                self._participant_id, message['id']), message)

            # Schedule the transition for processing
            task = helixtask.HelixTask(message, state_model, self._participant)
            self._threadpool.submit(task.call)



import unittest

import pyhelix.accessor as accessor
import pyhelix.helixexec as helixexec
import pyhelix.participant as participant
import pyhelix.statemodel as statemodel
import pyhelix.znode as znode

import mockclient


class MockStateModel(statemodel.StateModel):
    """
    A state model with nop transitions
    """
    def default_transition_handler(self, message):
        pass


class MockStateModelFactory(statemodel.StateModel):
    """
    A factory for a nop state model
    """
    def create_state_model(self, partition_name):
        """
        Create a MockStateModel

        Args:
            partition_name: The partition

        Returns:
            MockStateModel
        """
        return MockStateModel()


class MockParticipant(participant.Participant):
    """
    A participant that can't do anything substantial
    """
    def __init__(self, cluster_id, host, port, zk_addrs):
      participant.Participant.__init__(
          self, cluster_id, host, port, zk_addrs)
      self._client = mockclient.MockKazooClient()
      self._accessor = accessor.DataAccessor(cluster_id, self._client)

    def connect(self):
        self._is_lost = False
        self._client.start()


class TestExecutor(unittest.TestCase):
    """
    These test methods ensure proper parsing/handling of messages.
    """
    def setUp(self):
        factories = {'OnlineOffline': MockStateModelFactory()}
        self._p = MockParticipant(
            'mockcluster', 'localhost', 1234, 'localhost:2181')
        self._p.connect()
        self.executor = helixexec.HelixExecutor(factories, self._p)

    def test_session_id_mismatch(self):
        """
        Test that the executor removes messages with incorrect session ID
        """
        message = znode.get_empty_znode('MY_MESSAGE_ID')
        message['simpleFields']['MSG_TYPE'] = 'STATE_TRANSITION'
        message['simpleFields']['TGT_SESSION_ID'] = 'wrong'
        accessor = self._p.get_accessor()
        keybuilder = accessor.get_key_builder()
        participant_id = self._p.get_participant_id()
        message_id = message['id']
        message_key = keybuilder.message(participant_id, message_id)
        accessor.create(message_key, message)
        self.executor.on_message([message])
        self.assertFalse(accessor.exists(message_key))



import argparse
import logging
import threading
import time

import pyhelix.spectator as spectator

logging.basicConfig(level=logging.WARN)

# This is a dummy spectator. It connects, looks at the external view for the
# given partitions, and then exits.


def main(args):
    """
    This starts a spectator, uses it to look up partitions, then exits.
    """
    conn = spectator.SpectatorConnection(args.cluster, args.zkSvr)
    try:
        conn.connect()
        s = conn.spectate(args.resource)
        for partition_id in args.partitions:
            print s.get_state_map(partition_id)
    finally:
        conn.disconnect()

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--cluster', required=True, type=str, help='Name of the cluster')
    parser.add_argument(
        '--resource', required=True, type=str, help='Name of the resource')
    parser.add_argument(
        '--partitions', required=True, type=str, nargs='+',
        help='Partitions to look at')
    parser.add_argument(
        '--zkSvr', required=True, type=str, help='host:port of ZooKeeper')
    main(parser.parse_args())



import argparse
import bottle
import logging
import os
import random
import re
import urllib
import urllib2

import pyhelix.spectator as spectator


class CodeRunner(object):
    """
    A class that will find nodes that run code and dispatch work
    """
    def __init__(self, cluster, resource, host, port, zk_svr):
        """
        Initialize the CodeRunner

        Args:
            cluster: the cluster id
            resource: the resource id
            host: the hostname that this router can be reached on
            port: the port to use to send data
            zk_svr: host:port of a ZooKeeper server
        """
        self._conn = spectator.SpectatorConnection(cluster, zk_svr)
        self._s = None
        self._resource = resource
        self._app = bottle.Bottle()
        self._host = host
        self._port = port
        self._route()

    def start(self):
        """
        Start the CodeRunner
        """
        self._conn.connect()
        self._s = self._conn.spectate(self._resource)
        self._app.run(host='0.0.0.0', port=self._port)

    def stop(self):
        """
        Stop the CodeRunner
        """
        self._conn.disconnect()

    def show_index(self):
        """
        Called when the home page is requested

        Returns:
            A page with a box to input code where to run it
        """
        participants = self._s.get_participants('ONLINE')
        participants = [p['id'] for p in participants]
        return bottle.template(
            'index', host=self._host, port=self._port,
            participants=participants)

    def run_program(self):
        """
        Called when a request with a program to run comes in

        Returns:
            stdout result of running the program
        """
        prog = str(bottle.request.forms.get('prog'))
        node = str(bottle.request.forms.get('participant'))
        logging.info('Selected participant(s): {0}'.format(node))
        pattern = node
        if node == 'star':
            pattern = '.*'
        elif node == 'random':
            participants = self._s.get_participants('ONLINE')
            if len(participants) > 0:
                pattern = random.sample(participants, 1)[0]['id']
        outputs = self._run_on_nodes(prog, pattern)
        return bottle.template('result', results=outputs, prog=prog)

    def static_files(self, filename):
        """
        Static route to files

        Args:
            filename: name of the static file

        Returns:
            a file handle
        """
        path = os.path.dirname(os.path.realpath(__file__)) + '/static'
        return bottle.static_file(filename, root=path)

    def _run_on_nodes(self, prog, pattern):
        """
        Run a program on all machines that match a pattern.

        Args:
            prog: The text of the program
            pattern: The machine pattern

        Returns:
            The aggregate output
        """
        outputs = []
        pattern = re.compile(pattern, re.IGNORECASE)
        participants = self._s.get_participants('ONLINE')
        for participant in participants:
            if pattern.match(participant['id']):
                label = participant['id']
                host = participant['simpleFields']['HELIX_HOST']
                port = participant['simpleFields']['HELIX_PORT']
                values = {'prog': prog}
                data = urllib.urlencode(values)
                result = urllib2.urlopen(
                    'http://{0}:{1}/run'.format(host, port), data)
                output = label, result.read()
                outputs.append(output)
        return outputs

    def _route(self):
        """
        Route HTTP requests (private)
        """
        self._app.route('/', callback=self.show_index)
        self._app.route('/run', method='POST', callback=self.run_program)
        self._app.route('/static/<filename>', callback=self.static_files)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--zkSvr', required=True, type=str, help='host:port of ZooKeeper')
    parser.add_argument('--host', required=True, type=str, help='hostname')
    parser.add_argument('--port', required=True, type=str, help='port')
    args = parser.parse_args()
    r = CodeRunner(
        'coderunner-cluster', 'coderunner', args.host, args.port, args.zkSvr)
    r.start()
    r.stop()



import logging


class StateModelParser(object):
    """
    State model helper
    """

    def get_method_for_transition(self, clazz, from_state, to_state):
        """
        Get the callback for a state get_method_for_transition

        Args:
            clazz: The class that extends StateModel with the callbacks
            from_state: The previous state
            to_state: The new state

        Returns:
            The method to call
        """
        method_name = (
            'on_become_' + to_state.lower() + '_from_' + from_state.lower())
        logging.debug('method_name: {0}'.format(method_name))
        return getattr(clazz, method_name, clazz.default_transition_handler)


class StateModel(object):
    """
    Base state model
    """

    DEFAULT_INIT_STATE = 'OFFLINE'

    def __init__(self):
        """
        Initialize the basic state model
        """
        self._current_state = StateModel.DEFAULT_INIT_STATE

    def get_current_state(self):
        """
        Get the current state for this partition

        Returns:
            Current state string for this partition
        """
        return self._current_state

    def default_transition_handler(self, message):
        """
        Default method for when no method is available to handle the transition

        Args:
            message: the transition message
        """
        from_state = message['simpleFields']['FROM_STATE']
        to_state = message['simpleFields']['TO_STATE']
        logging.warn(
            'No method found for {0}-{1}'.format(from_state, to_state))


class StateModelFactory(object):
    """
    Base state model factory
    """

    def __init__(self):
        """
        Initialize the factory
        """
        self._state_models = {}

    def create_state_model(self, partition_name):
        """
        create a new state model (implemented by subclasses)
        """
        pass

    def put_state_model(self, partition_name, state_model):
        """
        Associate a partition with a state model

        Args:
            partition_name: The partition
            state_model: The state model
        """
        self._state_models[partition_name] = state_model

    def get_state_model(self, partition_name):
        """
        Get the associated state model for this partition

        Args:
            partition_name: The partition

        Returns:
            The state model, or None
        """
        if partition_name in self._state_models:
            return self._state_models[partition_name]
        return None



import argparse
import logging
import threading

import pyhelix.participant as participant

import dummy_statemodel

logging.basicConfig(level=logging.WARN)

# This is a dummy participant. It starts up, runs forever, and prints state
# transitions as it gets them. It's close to what a real participant would look
# like, except that it uses a DummyStateModel.


def main(args):
    p = participant.Participant(
        args.cluster, args.host, args.port, args.zkSvr,
        participant_id=args.participantId)
    try:
        p.register_state_model_fty(
            'MasterSlave', dummy_statemodel.DummyStateModelFactory())
        p.connect()

        # wait forever
        dummy_event = threading.Event()
        dummy_event.wait()
    finally:
        p.disconnect()

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--cluster', required=True, type=str, help='Name of the cluster')
    parser.add_argument(
        '--zkSvr', required=True, type=str, help='host:port of ZooKeeper')
    parser.add_argument('--host', required=True, type=str, help='hostname')
    parser.add_argument('--port', required=True, type=str, help='port')
    parser.add_argument(
        '--participantId', required=False, type=str,
        help="Optional custom participant ID")
    main(parser.parse_args())



import unittest
import kazoo.exceptions

import mockclient


class TestMockKazooClient(unittest.TestCase):
    """
    MockKazooClient is nontrivial, so this verifies basic correctness
    """

    def setUp(self):
        self.c = mockclient.MockKazooClient()
        self.c.start()

    def test_connected(self):
        """
        Connection state should correspond to start/stop
        """
        self.assertTrue(self.c.connected)
        self.c.stop()
        self.assertFalse(self.c.connected)

    def test_create_basic(self):
        """
        Test that a default creation is successful
        """
        path = '/testPath'
        result = self.c.create(path, 'testData')
        self.assertEqual(result, path)
        data, metadata = self.c.get(path)
        self.assertTrue(metadata is not None)
        self.assertEqual(data, 'testData')

    def test_create_recursive(self):
        """
        Test that recursive creation only passes if makepath is True
        """
        # empty set
        path = '/path/testPath'
        self.assertRaises(
            kazoo.exceptions.NoNodeError, self.c.create, path, 'testData')
        result = self.c.create(path, 'testData', makepath=True)
        self.assertEqual(result, path)

        # middle is missing
        path2 = path + '/two/more'
        self.assertRaises(
            kazoo.exceptions.NoNodeError, self.c.create, path2, 'testData2')
        result = self.c.create(path2, 'testData2', makepath=True)
        self.assertEqual(result, path2)

    def test_create_ephemeral(self):
        """
        Test ephemeral node creation behavior
        """
        path = '/testPath'
        result = self.c.create(path, 'testData', ephemeral=True)
        self.assertTrue(result, path)

        # it's not allowed to have a child of an ephemeral node
        path = path + '/child'
        self.assertRaises(
            kazoo.exceptions.NoChildrenForEphemeralsError, self.c.create, path,
            'testChildData')

    def test_ensure_path(self):
        """
        Ensure path should create the path and all subpaths
        """
        # this shouldn't raise any errors
        path = '/one/two/three/four'
        self.c.ensure_path(path)
        data, metadata = self.c.get('/one')
        self.assertTrue(metadata is not None)
        data, metadata = self.c.get('/one/two')
        self.assertTrue(metadata is not None)
        data, metadata = self.c.get('/one/two/three')
        self.assertTrue(metadata is not None)
        data, metadata = self.c.get('/one/two/three/four')
        self.assertTrue(metadata is not None)
        self.assertTrue(self.c.exists('/one'))
        self.assertTrue(self.c.exists('/one/two'))
        self.assertTrue(self.c.exists('/one/two/three'))
        self.assertTrue(self.c.exists('/one/two/three/four'))

    def test_exists(self):
        """
        Ensure that exists is correct
        """
        path = '/one/two/three/four'
        self.assertFalse(self.c.exists('/one'))
        self.c.create(path, 'testData', makepath=True)
        self.assertTrue(self.c.exists('/one'))
        self.assertTrue(self.c.exists('/one/two'))
        self.assertTrue(self.c.exists('/one/two/three'))
        self.assertTrue(self.c.exists('/one/two/three/four'))

    def test_get_nonexistent(self):
        """
        Ensure that get raises an error if the node doesn't exist
        """
        path = '/one/two/three/four'
        self.assertRaises(kazoo.exceptions.NoNodeError, self.c.get, path)

    def test_get_children(self):
        """
        Test that get children gets all children and no others
        """
        self.c.create('/one', 'onedata')
        self.c.create('/outside', 'outsidedata')
        self.c.create('/outside/a', 'outsideadata')
        self.c.create('/one/two', 'twodata')
        self.c.create('/one/three', 'threedata')
        children = self.c.get_children('/one')
        self.assertTrue('/one/two' in children)
        self.assertTrue('/one/three' in children)
        self.assertTrue('/one' not in children)
        self.assertTrue('/outside' not in children)
        self.assertTrue('/outside/a' not in children)
        self.assertEqual(len(children), 2)

    def test_basic_set(self):
        """
        Test that set updates the value
        """
        path = '/one'
        self.assertRaises(
            kazoo.exceptions.NoNodeError, self.c.set, path, 'data')
        self.c.create(path, 'data')
        self.c.set(path, 'updated')
        data, metadata = self.c.get(path)
        self.assertEqual(data, 'updated')
        self.assertTrue(metadata is not None)

    def test_set_version_mismatch(self):
        """
        Test that set behaves correctly on version match/mismatch
        """
        path = '/one'
        self.c.create(path, 'data')
        self.c.set(path, 'updated', version=1)
        data, metadata = self.c.get(path)
        self.assertEqual(data, 'updated')
        self.c.set(path, 'updated2', version=2)
        data, metadata = self.c.get(path)
        self.assertEqual(data, 'updated2')
        self.assertRaises(
            kazoo.exceptions.BadVersionError, self.c.set, path, 'updated3',
            version=1)
        data, metadata = self.c.get(path)
        self.assertNotEqual(data, 'updated3')
        self.c.set(path, 'updated4', version=-1)
        data, metadata = self.c.get(path)
        self.assertEqual(data, 'updated4')

    def test_basic_delete(self):
        """
        Test that a node is removed once delete is called
        """
        path = '/one'
        self.assertRaises(kazoo.exceptions.NoNodeError, self.c.delete, path)
        self.c.create(path, 'data')
        data, metadata = self.c.get(path)
        self.assertEqual(data, 'data')
        self.c.delete(path)
        self.assertFalse(self.c.exists(path))

    def test_recursive_delete(self):
        """
        Test that recursive delete does the right thing
        """
        self.c.create('/one', None)
        self.c.create('/one/a', None)
        self.c.create('/one/b', None)
        self.c.create('/two', None)
        self.c.create('/two/a', None)
        self.assertRaises(
            kazoo.exceptions.NotEmptyError, self.c.delete, '/one')
        self.c.delete('/one', recursive=True)
        self.assertFalse(self.c.exists('/one'))
        self.assertFalse(self.c.exists('/one/a'))
        self.assertFalse(self.c.exists('/one/b'))
        self.assertTrue(self.c.exists('/two'))
        self.assertTrue(self.c.exists('/two/a'))

    def test_delete_version_mismatch(self):
        """
        Test that delete pays attention to version
        """
        path = '/one'
        self.c.create(path, 'data')
        self.c.set(path, 'updated', version=2)
        self.assertRaises(
            kazoo.exceptions.BadVersionError, self.c.delete, path, version=1)
        self.c.delete(path, version=3)
        self.assertFalse(self.c.exists(path))
        self.c.create(path, 'data')
        self.c.set(path, 'updated', version=2)
        self.c.delete(path, version=3)

    def tearDown(self):
        self.c.stop()
        self.c.close()



import setuptools

import pyhelix.constants as constants


def readme():
    with open('README.md') as f:
        return f.read()

setuptools.setup(
    name='pyhelix',
    version=str(constants.CURRENT_VERSION),
    description='Python bindings for Apache Helix',
    long_description=('Support for creating Helix participants in Python.'
                      ' Register a state model, connect, and go!'),
    url='http://github.com/kanakb/pyhelix',
    license='Apache License 2.0',
    packages=setuptools.find_packages(),
    keywords='helix',
    install_requires=['kazoo', 'futures', 'argparse'],
    include_package_data=True,
    test_suite='nose.collector',
    tests_require=['nose'],
    author='Kanak Biscuitwala, Zhen Zhang',
    author_email='kanak.b@hotmail.com, nehzgnahz@gmail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.6',
        ],
    zip_safe=False)



import logging
import threading

import statemodel
import znode


class HelixTask(object):
    """
    Helix task for state transitions
    """

    def __init__(self, message, state_model, participant):
        """
        Instantiate this task

        Args:
            message: The message to be processed
            state_model: The state model that the participant follows
            participant: Participant connection
        """
        self._message = message
        self._state_model = state_model
        self._state_model_parser = statemodel.StateModelParser()
        self._accessor = participant.get_accessor()
        self._builder = self._accessor.get_key_builder()
        self._participant_id = participant.get_participant_id()
        self._participant = participant

    def call(self):
        """
        Process the message.

        Calls application callbacks and then updates the global state.
        """
        thread = threading.current_thread()
        logging.info('{0} invokes message: {1}'.format(
            str(thread), self._message))
        from_state = self._message['simpleFields']['FROM_STATE']
        to_state = self._message['simpleFields']['TO_STATE']
        method_to_invoke = self._state_model_parser.get_method_for_transition(
            self._state_model, from_state, to_state)
        logging.info('method_to_invoke: {0}'.format(str(method_to_invoke)))
        method_to_invoke(self._message)
        self._state_model._current_state = to_state

        # update current-state, then remove message
        sub = False
        resource_name = self._message['simpleFields']['RESOURCE_NAME']
        partition_name = self._message['simpleFields']['PARTITION_NAME']
        current_state = znode.get_empty_znode(resource_name)
        current_state['mapFields'][partition_name] = {}
        if to_state != 'DROPPED':
            # update the current state
            current_state['mapFields'][partition_name]['CURRENT_STATE'] = (
                to_state)
            current_state['simpleFields']['STATE_MODEL_DEF'] = (
                self._message['simpleFields']['STATE_MODEL_DEF'])
            current_state['simpleFields']['SESSION_ID'] = (
                self._participant.get_session_id())
        else:
            # drop the partition from the current state
            sub = True
        self._accessor.update(self._builder.current_state(
            self._participant_id, self._participant.get_session_id(),
            resource_name), current_state, sub=sub)
        self._accessor.remove(self._builder.message(
            self._participant_id, self._message['id']))



import json
import kazoo.exceptions
import kazoo.recipe.watchers
import logging
import os
import traceback

import keybuilder


class DataAccessor(object):
    """
    Helix property accessor
    """
    def __init__(self, cluster_id, zk_client):
        """
        Initialze for a cluster and ZooKeeper connection

        Args:
            cluster_id: Unique clsuter identifier
            zk_client: A live connection to ZooKeeper
        """
        self._cluster_id = cluster_id
        self._client = zk_client

    def create(self, key, data):
        """
        Create a property, creating parent nodes as necessary

        Args:
            key: KeyBuilder property
            data: The data to write (python object)

        Returns:
            True if successful, False otherwise
        """
        path = key['path']
        try:
            node = data
            if data:
                node = json.dumps(data, indent=2, sort_keys=True)
            logging.info('creating {0} with {1}'.format(path, data))
            self._client.create(
                path, node, ephemeral=key['ephemeral'],
                sequence=key['sequential'], makepath=True)
            return True
        except kazoo.exceptions.NodeExistsError:
            logging.warn('{0} exists already'.format(path))
            return self.set(key, data)
        except kazoo.exceptions.KazooException:
            logging.error(path)
            logging.error(traceback.format_exc())
        return False

    def set(self, key, data):
        """
        Set a property, creating parent nodes as necessary

        Args:
            key: KeyBuilder property
            data: The data to write (python object)

        Returns:
            True if successful, False otherwise
        """
        if data:
            data = json.dumps(data, indent=2, sort_keys=True)
        path = key['path']
        try:
            if not key['update_only_on_exists']:
                self._client.ensure_path(path)
            logging.info('setting {0} with {1}'.format(path, data))
            self._client.set(path, data)
            return True
        except kazoo.exceptions.NoNodeError:
            logging.info('{0} does not exist'.format(path))
        except kazoo.exceptions.KazooException:
            logging.error(path)
            logging.error(traceback.format_exc())
        return False

    def get(self, key):
        """
        Get a property

        Args:
            key: KeyBuilder property

        Returns:
            The data that is persisted, or None
        """
        path = key['path']
        try:
            value, stat = self._client.get(path)
            return json.loads(value)
        except kazoo.exceptions.NoNodeError:
            logging.info('{0} does not exist'.format(path))
        except kazoo.exceptions.KazooException:
            logging.error(path)
            logging.error(traceback.format_exc())
        return None

    def get_children(self, key):
        """
        Get the children of a property

        Args:
            key: KeyBuilder property

        Returns:
            List of child names
        """
        path = key['path']
        try:
            return self._client.get_children(path)
        except kazoo.exceptions.NoNodeError:
            logging.info('{0} does not exist'.format(path))
        except kazoo.exceptions.KazooException:
            logging.error(path)
            logging.error(traceback.format_exc())
        return []

    def update(self, key, updated_value, sub=False):
        """
        Update a property

        Args:
            key: KeyBuilder property
            updated_value: The data to write (python object)
            sub: True to subtract the updated value, False (default) to add it

        Returns:
            True if successful, False otherwise
        """
        path = key['path']
        done = False
        while not done:
            try:
                exists_stat = self._client.exists(path)
                if not exists_stat:
                    if key['update_only_on_exists'] or sub:
                        logging.info(
                            '{0} does not exist, cannot update'.format(path))
                        return False
                    try:
                        node = updated_value
                        if updated_value:
                            node = json.dumps(updated_value, indent=2, sort_keys=True)
                        self._client.create(
                            path, node, ephemeral=key['ephemeral'],
                            sequence=key['sequential'], makepath=True)
                        return True
                    except kazoo.exceptions.NodeExistsError:
                        pass  # ignore failure here
                value, get_stat = self._client.get(path)
                value = json.loads(value)
                if key['merge_on_update']:
                    # merge in if allowed
                    if not sub:
                        for k, v in updated_value['simpleFields'].iteritems():
                            value['simpleFields'][k] = v
                        for k, v in updated_value['listFields'].iteritems():
                            value['listFields'][k] = v
                        for k, v in updated_value['mapFields'].iteritems():
                            value['mapFields'][k] = v
                    else:
                        for k, v in updated_value['simpleFields'].iteritems():
                            if k in value['simpleFields']:
                                value['simpleFields'].pop(k)
                        for k, v in updated_value['listFields'].iteritems():
                            if k in value['listFields']:
                                value['listFields'].pop(k)
                        for k, v in updated_value['mapFields'].iteritems():
                            if k in value['mapFields']:
                                value['mapFields'].pop(k)
                elif not sub:
                    # otherwise, replace entirely if not a subtraction
                    value = updated_value
                else:
                    # merge must be allowed to subtract
                    logging.warn(
                        'Tried to do a subtract on a property that'
                        ' doesn\'t allow merge')
                    return False
                value = json.dumps(value, indent=2, sort_keys=True)
                update_stat = self._client.set(
                    path, value, version=get_stat.version)
                if update_stat:
                    done = True
            except kazoo.exceptions.BadVersionError:
                logging.info('trying again to update {0}'.format(path))
                continue  # ignore this, try again
            except kazoo.exceptions.KazooException:
                logging.error(path)
                logging.error(traceback.format_exc())
                return False
        return True

    def remove(self, key):
        """
        Remove a property

        Args:
            key: KeyBuilder property

        Returns:
            True if successful, False otherwise
        """
        path = key['path']
        try:
            self._client.delete(path, recursive=True)
            return True
        except kazoo.exceptions.NoNodeError:
            logging.warn('{0} does not exist'.format(path))
        except kazoo.exceptions.KazooException:
            logging.error(path)
            logging.error(traceback.format_exc())
        return False

    def exists(self, key):
        """
        Check if a property exists

        Args:
            key: KeyBuilder property

        Returns:
            True if exists, False otherwise
        """
        path = key['path']
        try:
            return self._client.exists(path) is not None
        except kazoo.exceptions.KazooException:
            logging.error(path)
            logging.error(traceback.format_exc())
        return False

    def watch_children(self, key, func):
        """
        Watch children of a property

        Args:
            key: KeyBuilder property
            func: Single argument callback
        """
        path = key['path']
        kazoo.recipe.watchers.ChildrenWatch(self._client, path, func=func)

    def watch_property(self, key, func):
        """
        Watch a property

        Args:
            key: KeyBuilder property
            func: Two-argument callback that takes data, stat
        """
        path = key['path']
        kazoo.recipe.watchers.DataWatch(self._client, path, func=func)

    def get_key_builder(self):
        """
        Get a key builder that can be used with this accessor.

        Returns:
            KeyBuilder instance
        """
        return keybuilder.KeyBuilder(self._cluster_id)
